@@ -33,6 +33,8 @@ import dbus
 from dbus.mainloop.glib import DBusGMainLoop
 DBusGMainLoop(set_as_default=True)
 
+from beye.sliceset import SliceSet as SliceSet
+
 PERCENTAGE = beye.DataSourceHal.PERCENTAGE
 CHARGE = beye.DataSourceHal.CHARGE
 VOLTAGE = beye.DataSourceHal.VOLTAGE
@@ -174,9 +176,8 @@ class MainWindow(hildon.StackableWindow):
 
     def run(self):
         self.show_all()
-        self.graph.jump_to(self.graph.drawAreaInfo.canvasSize[0]-1, 0)
-        gobject.idle_add(self.firstIdleCb)
         hildon.hildon_gtk_window_set_progress_indicator(self, True)
+        gobject.idle_add(self.firstIdleCb)
         bus = dbus.SystemBus()
         bus.add_signal_receiver(self.dbusHandler, path='/org/freedesktop/Hal/devices/bme')
         try:
@@ -186,9 +187,7 @@ class MainWindow(hildon.StackableWindow):
             print "Ctrl-C"
 
     def firstIdleCb(self):
-        self.graph.maxVisibleHours = 30*24
-        self.graph.update()
-        hildon.hildon_gtk_window_set_progress_indicator(self, False)
+        self.doUpdate()
         return False
 
     def queueUpdate(self):
@@ -246,6 +245,8 @@ class MainWindow(hildon.StackableWindow):
         self.set_app_menu(None)
         gtk.main_quit()
 
+
+
 class DrawAreaInfo(object):
     def __init__(self, canvasSize, drawAreaSize, offset, windowSize):
         self.canvasSize = canvasSize
@@ -253,6 +254,12 @@ class DrawAreaInfo(object):
         self.offset = offset
         self.windowSize = windowSize
 
+    def canvasAsRect(self):
+        return gtk.gdk.Rectangle(0, 0, self.canvasSize[0], self.canvasSize[1])
+
+    def drawAreaAsRect(self):
+        return gtk.gdk.Rectangle(self.offset[0], self.offset[1], self.drawAreaSize[0], self.drawAreaSize[1])
+
 class EraseableBuffer(object):
     def __init__(self, pixmap):
         self.pixmap = pixmap
@@ -302,6 +309,7 @@ class Graph(hildon.PannableArea):
         self.area.set_double_buffered(False)
         self.set_double_buffered(False)
         self.add_with_viewport(self.area)
+        self.pixmap = None
 
         self.area.connect("expose-event", self.redrawGraph)
 
@@ -329,19 +337,17 @@ class Graph(hildon.PannableArea):
                               CHARGE: [],
                               PERCENTAGE: []}
 
+        self.unrenderedArea = SliceSet()
+
         self.chargeInfo = []
 
         self.visibleHours = int(self.dataStorage.getConfig('display:visibleHours', 25))
-        self.maxVisibleHours = self.visibleHours
+        self.maxVisibleHours = 30*24
 
         self.gridType = self.dataStorage.getConfig('display:gridType', PERCENTAGE)
 
         self.requestJump = None
 
-        self.updateGrid()
-        self.updateData()
-        self.renderData()
-
     def update(self):
         self.dataRenderers = {VOLTAGE: [],
                               CHARGE: [],
@@ -349,12 +355,17 @@ class Graph(hildon.PannableArea):
         self.chargeInfo = []
         self.updateGrid()
         self.updateData()
-        self.renderData()
+        #self.renderData(self.drawAreaInfo.canvasAsRect())
         self.requestJump = self.drawAreaInfo.canvasSize[0]-1
+        self.area.queue_draw()
+        #gobject.idle_add(self.moveLabels)
 
     def updateIncremental(self):
         self.updateDataIncremental()
-        self.renderData()
+        self.unrenderedArea = SliceSet([(0, self.drawAreaInfo.canvasSize[0] - 1)])
+        self.area.queue_draw()
+        #self.renderData(self.drawAreaInfo.canvasAsRect())
+        #gobject.idle_add(self.moveLabels)
 
     def enableDataset(self, dataset):
         #print "Enable: %s" % dataset
@@ -362,7 +373,10 @@ class Graph(hildon.PannableArea):
         self.dataStorage.setConfig('display:' + dataset, 1)
         self.updateData()
         self.changeYLabels(dataset)
-        self.renderData()
+        self.unrenderedArea = SliceSet([(0, self.drawAreaInfo.canvasSize[0] - 1)])
+        self.area.queue_draw()
+        #self.renderData(self.drawAreaInfo.canvasAsRect())
+        #gobject.idle_add(self.moveLabels)
 
     def disableDataset(self, dataset):
         #print "Disable: %s" % dataset
@@ -370,7 +384,10 @@ class Graph(hildon.PannableArea):
         self.dataStorage.setConfig('display:' + dataset, 0)
         if self.gridType == dataset:
             self.changeYLabels(self.getBestYLabelType())
-        self.renderData()
+        self.unrenderedArea = SliceSet([(0, self.drawAreaInfo.canvasSize[0] - 1)])
+        self.area.queue_draw()
+        #self.renderData(self.drawAreaInfo.canvasAsRect())
+        #gobject.idle_add(self.moveLabels)
 
     def changeVisibleHours(self, visibleHours):
         #print "Visible Hours: %d" % visibleHours
@@ -383,8 +400,11 @@ class Graph(hildon.PannableArea):
                 r.changeScale(self.drawAreaInfo, timeBounds)
         for r in self.chargeInfo:
             r.changeScale(self.drawAreaInfo, timeBounds)
-        self.renderData()
+        #self.renderData(self.drawAreaInfo.canvasAsRect())
         self.requestJump = self.drawAreaInfo.canvasSize[0]-1
+        self.area.queue_draw()
+        #gobject.idle_add(self.moveLabels)
+        
 
     def getBestYLabelType(self):
         for s in (PERCENTAGE, CHARGE, VOLTAGE):
@@ -469,6 +489,7 @@ class Graph(hildon.PannableArea):
         
         self.pixmap = gtk.gdk.Pixmap(None, areaWidth, self.graphHeight, 16)
         self.area.set_size_request(areaWidth, self.graphHeight)
+        self.unrenderedArea = SliceSet([(0, self.drawAreaInfo.canvasSize[0] - 1)])
         
     def updateData(self):
         timeBounds = (self.xMin, self.xMax)
@@ -525,19 +546,25 @@ class Graph(hildon.PannableArea):
             for segment in [s for s in rawData if len(rawData)]:
                 self.dataRenderers[type].append(DataRendererValue(segment, self.drawAreaInfo, timeBounds, valueBounds))
 
-    def renderData(self):
+    def renderData(self, cliprect):
         gc = self.pixmap.new_gc()
-
+        gc.set_clip_rectangle(cliprect)
         gc.foreground = self.bgColor
+
+        area = self.grid.clearYLabels(self.pixmap)
+        if area:
+            self.area.queue_draw_area(area.x, area.y, area.width, area.height)
+
         self.pixmap.draw_rectangle(gc, True,
                                    0, 0,
                                    self.drawAreaInfo.canvasSize[0],
                                    self.drawAreaInfo.canvasSize[1])
-
+        
         for d in self.chargeInfo:
-            d.render(self.pixmap, {(True, True): self.fullColor, (True, False): self.chargingColor})
+            d.render(self.pixmap, {(True, True): self.fullColor, (True, False): self.chargingColor},
+                     cliprect)
 
-        self.grid.render(self.pixmap, self.gridColor)
+        self.grid.render(self.pixmap, self.gridColor, cliprect)
 
         for type, renderers in ((t, self.dataRenderers[t])
                                 for t in (VOLTAGE,
@@ -546,28 +573,50 @@ class Graph(hildon.PannableArea):
             #print "%s: %d sets" % (type, len(renderers))
             c = self.dataColors[type]
             for d in renderers:
-                d.render(self.pixmap, c)
+                d.render(self.pixmap, c, cliprect)
 
-        self.grid.drawYLabels(self.pixmap, int(self.get_hadjustment().get_value()))
         self.area.queue_draw()
-
+        self.moveLabels(False)
+            
     def redrawGraph(self, widget, event):
-        self.gc = widget.style.fg_gc[gtk.STATE_NORMAL]
-
-        area = event.area
-
-        #print "Redraw (%d, %d)->(%d, %d)" % (area.x, area.y, area.x + area.width, area.y + area.height)
-
-        widget.window.draw_drawable(self.gc, self.pixmap,
-                                    area.x, area.y,
-                                    area.x, area.y,
-                                    area.width, area.height)
+        if not self.pixmap:
+            # We're not currently initialized, ignore.
+            return
+            
         if (self.requestJump != None):
             self.jump_to(self.requestJump, 0)
             self.requestJump = None
+            self.area.queue_draw()
+            return
+        
+        self.gc = widget.style.fg_gc[gtk.STATE_NORMAL]
+            
+        area = event.area
 
-    def moveLabels(self, adjustment):
-        area = self.grid.drawYLabels(self.pixmap, int(self.get_hadjustment().get_value()))
+#        print "Redraw (%d, %d)->(%d, %d)" % (area.x, area.y, area.x + area.width, area.y + area.height)
+
+        edges = ()
+        if self.unrenderedArea.intersection((area.x, area.x + area.width - 1)):
+            for i in self.unrenderedArea.intersection((area.x - self.drawAreaInfo.windowSize[0],
+                                                       area.x + area.width + self.drawAreaInfo.windowSize[0])):
+                edges += i
+
+        if len(edges):
+            x = min(edges)
+            w = max(edges) - min(edges) + 1
+            r = gtk.gdk.Rectangle(x, 0, w, self.drawAreaInfo.canvasSize[1])
+            #print "Render (%d, %d)->(%d, %d)" % (r.x, r.y, r.x + r.width, r.y + r.height)
+            self.unrenderedArea.subtract((x, max(edges)))
+            gobject.idle_add(self.renderData, r)
+        else:
+            widget.window.draw_drawable(self.gc, self.pixmap,
+                                        area.x, area.y,
+                                        area.x, area.y,
+                                        area.width, area.height)
+
+    def moveLabels(self, adjustment=False):
+        #print "Draw Y labels: %d" % int(self.get_hadjustment().get_value())
+        area = self.grid.drawYLabels(self.pixmap, int(self.get_hadjustment().get_value())) 
         self.area.queue_draw_area(area.x, area.y, area.width, area.height)
 
     def getChargeInfo(self, intervalStart, intervalEnd):
@@ -808,8 +857,9 @@ class GridRenderer(object):
             return [(x0, p, x1, p) for p in
                         (int(scale * (p - bounds[0])) + dispY for p in points)]
 
-    def render(self, drawable, gridColor):
+    def render(self, drawable, gridColor, cliprect):
         gc = drawable.new_gc(foreground=gridColor, line_width=1, function=gtk.gdk.XOR)
+        gc.set_clip_rectangle(cliprect)
         drawable.draw_segments(gc, self.yGrid)
         drawable.draw_segments(gc, self.xGrid)
         self.drawXLabels(gc, drawable, self.xLabels)
@@ -821,6 +871,7 @@ class DataRendererValue(object):
         self.timeBounds = timeBounds
         self.valueBounds = valueBounds
         self.scaledData = self.rescaleData(self.rawData, self.areaInfo, self.timeBounds, self.valueBounds)
+        self.coordBounds = (self.scaledData[0][0]-1, self.scaledData[-1][0]+1)
 
     def getLastObservationTime(self):
         if len(self.rawData):
@@ -830,6 +881,7 @@ class DataRendererValue(object):
     def addPoints(self, points):
         self.rawData[len(self.rawData):] = points
         self.scaledData[len(self.scaledData):] = self.rescaleData(points, self.areaInfo, self.timeBounds, self.valueBounds)
+        self.coordBounds = (self.scaledData[0][0]-1, self.scaledData[-1][0]+1)
 
     def changeScale(self, areaInfo, timeBounds, valueBounds=None):
         self.areaInfo = areaInfo
@@ -856,12 +908,17 @@ class DataRendererValue(object):
         return [(int(scaleX * (p[0] - xBounds[0])) + offX,
                  int(scaleY * (p[1] - yBounds[0])) + dispY) for p in points]
 
-    def render(self, drawable, color):
+    def render(self, drawable, color, cliprect):
+        if max(cliprect.x, self.coordBounds[0]) > min(cliprect.x + cliprect.width - 1, self.coordBounds[1]):
+            # Nothing to see here, move along.
+            return
         gc = drawable.new_gc(foreground=color, line_width=1, join_style=gtk.gdk.JOIN_ROUND)
+        gc.set_clip_rectangle(cliprect)
         color2 = drawable.get_colormap().alloc_color(int(color.red*0.5),
                                                      int(color.green*0.5),
                                                      int(color.blue*0.5))
         gc2 = drawable.new_gc(foreground=color2, line_width=2, join_style=gtk.gdk.JOIN_ROUND)
+        gc2.set_clip_rectangle(cliprect)
 
         if len(self.scaledData) == 1:
             drawable.draw_points(gc2, self.scaledData)
@@ -876,6 +933,7 @@ class DataRendererStatus(object):
         self.areaInfo = areaInfo
         self.timeBounds = timeBounds
         self.scaledData = self.rescaleData(self.rawData, self.areaInfo, self.timeBounds)
+        self.coordBounds = (self.scaledData[0][1], self.scaledData[-1][2])
 
     def getFirstObservationTime(self):
         if not len(self.rawData):
@@ -903,15 +961,21 @@ class DataRendererStatus(object):
                  int(scaleX * (s[1] - xBounds[0])) + offX,
                  int(scaleX * (s[2] - xBounds[0])) + offX) for s in statuses]
 
-    def render(self, drawable, colorMap):
+    def render(self, drawable, colorMap, cliprect):
+        if max(cliprect.x, self.coordBounds[0]) > min(cliprect.x + cliprect.width - 1, self.coordBounds[1]):
+            # Nothing to see here, move along.
+            return
+        #print "render -> %d - %d" % self.coordBounds
         for status in self.scaledData:
             if (not colorMap.has_key(status[0])) or colorMap[status[0]] == None:
                 continue
             gc = drawable.new_gc(foreground=colorMap[status[0]])
+            gc.set_clip_rectangle(cliprect)
             drawable.draw_rectangle(gc, True,
                                     status[1], self.areaInfo.offset[1],
                                     status[2] - status[1], self.areaInfo.offset[1] + self.areaInfo.drawAreaSize[1] - 14)
 
+
 def getHoursInInterval(start, end, modulo):
     hour = datetime.timedelta(0, 0, 0, 0, 0, 1)
     interval = datetime.timedelta(0, 0, 0, 0, 0, modulo)
